import os

import apium.frameworks


def test_basic_task_run___state_is_consistent(port_num, running_worker):
    os.environ['DJANGO_SETTINGS_MODULE'] = 'django_test_project.settings'
    apium.frameworks.setup()



import apium
import pytest
import time


def add(*args):
    time.sleep(0.1)
    return sum(args)


def raiser(*args):
    raise ValueError(args)


def format_exc(exc):
    return str(exc)


def is_port_open(port):
    import socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex(('127.0.0.1', port))
    return result == 0


@pytest.fixture(scope="session")
def port_num():
    import random

    def random_port():
        return random.randint(1025, 65535)

    port = random_port()
    while is_port_open(port):
        port = random_port()
    return port


@pytest.yield_fixture(scope="session")
def task_add():
    fn = apium.register_task(add)
    try:
        yield fn
    finally:
        del apium.worker.WorkersState.tasks['add']


@pytest.yield_fixture(scope="session")
def task_raise():
    fn = apium.register_task(raiser)
    try:
        yield fn
    finally:
        del apium.worker.WorkersState.tasks['raiser']


@pytest.yield_fixture(scope="session")
def task_format_exc():
    fn = apium.register_task(format_exc)
    try:
        yield fn
    finally:
        del apium.worker.WorkersState.tasks['format_exc']


@pytest.yield_fixture(scope="session")
def running_worker(port_num, task_add, task_raise, task_format_exc):
    import threading
    address = ('localhost', port_num)
    with apium.worker.create_workers(['task_import'], 2) as workers:
        with apium.worker.setup_tcp_server(address, workers) as tcp_server:
            with apium.worker.setup_scheduler(address, 0.01):
                thread = threading.Thread(target=tcp_server.serve_forever)
                thread.start()
                while not is_port_open(port_num):
                    time.sleep(0.1)
                yield



import apium


def chain(*iterables):
    result = []
    for it in iterables:
        for element in it:
            result.append(element)
    return result


apium.register_task(chain)


@apium.register_task
def scheduled_fn(filename):
    with open(filename, 'a') as f:
        f.write('1')



# -*- coding: utf-8 -*-

"""
.. module: worker
    :synopsis: Spawns the workers, scheduler thread, and creates the TCP server.
"""


__all__ = (
    'register_task',
    'schedule_task',
    'create_workers',
    'setup_scheduler',
    'setup_tcp_server',
)


import collections
import concurrent.futures
import contextlib
import functools
import importlib
import logging
import pickle
import threading
import time
import traceback
import uuid

from datetime import datetime
from multiprocessing import Queue, Lock

from .client import sendmsg
from .exceptions import (
    RemoteException,
    TaskDoesNotExist,
    TaskWasNotSubmitted,
    UnknownMessage,
)
from .utils import format_fn


try:
    from inspect import signature
except ImportError:
    from funcsigs import signature


try:
    import socketserver
except ImportError:
    import SocketServer as socketserver


class WorkersState(object):
    """Namespace in which to keep the module's state."""
    # TODO: Can we please make this thread-safe?!?!?
    tasks = {}
    futures = {}
    chains = collections.defaultdict(list)
    chains_lock = Lock()
    schedule_queue = Queue()
    schedules = collections.defaultdict(list)


def register_task(fn):
    """A decorator to register a function as a task, to be later run by a
    Worker.

    .. doctest::

        >>> import apium
        >>> @apium.register_task
        ... def task1(arg1, kwarg1=1):
        ...     return arg1 + kwarg1

    :param fn: The function to register.
    :type fn: function
    :returns: The given function.
    :rtype: function
    """
    fn_name = fn.__name__
    logging.debug('Registering task %s', fn_name)
    WorkersState.tasks[fn_name] = fn
    return fn


def schedule_task(start_when=None, repeat_every=None, args=None, kwargs=None):
    """A decorator to register and schedule the given function at a time and/or
    interval.

    It's possible to apply this decorator multiple times to the same function in
    order to have multiple schedules for the same function.

    .. testsetup::

        from datetime import datetime, timedelta

    .. doctest::

        >>> import apium
        >>> @apium.schedule_task(datetime.now() + timedelta(seconds=30), args=(2, ))
        ... @apium.schedule_task(repeat_every=timedelta(seconds=60), args=(3, ), kwargs={'kwarg1': 3})
        ... @apium.schedule_task(args=(4, ), kwargs={'kwarg1': 5})
        ... def task1(arg1, kwarg1=1):
        ...     return arg1 + kwarg1

    In the example above, task1 will be added to the worker's queue:

     - as task1(2, kwarg1=1), 30 seconds after the workers start.

     - as task1(3, kwarg1=3), immediately after the workers start and every 60 seconds thereafter.

     - as task1(4, kwarg1=5), immediately after the workers start.

    :param start_when: When to run the function for the first time. A value of
        None results in the function being run immediately after the workers
        start.
    :type start_when: datetime.datetime
    :param repeat_every: The interval at which the function should repeatedly
        run. A value of None results in the function only being called once.
    :type repeat_every: datetime.timedelta
    :param args: The arguments to pass to the scheduled function.
    :type args: tuple
    :param kwargs: The keyword arguments to pass to the scheduled function.
    :type kwargs: dict
    :returns: The given function
    :rtype: function
    """
    def _schedule_task(fn):
        fn_name = fn.__name__
        logging.debug(
            'Scheduling task %s for %s, repeating every %s',
            format_task({'name': fn_name, 'args': args, 'kwargs': kwargs}),
            start_when or datetime.now(),
            repeat_every
        )
        schedule = (start_when or datetime.now(), repeat_every, args or (), kwargs or {})
        WorkersState.schedule_queue.put((fn_name, schedule))
        return register_task(fn)
    return _schedule_task


def format_task(task, chain=False):
    """Return a string representation of a task, for logging."""
    if chain:
        task = task.copy()
        task['args'] = ('X', ) + task['args']

    return '{}{}{}'.format(
        format_fn(task['name'], task['args'], task['kwargs']),
        ' [{}]'.format(task['id'].decode()) if task.get('id') else '',
        ' from {}'.format(task['client']) if task.get('client') else '',
    )


def run_task(task):
    """A wrapper around a task's function, for handling logging and exceptions."""
    formatted_task = format_task(task)
    logging.info('Running task %s...', formatted_task)
    task_start = datetime.now()
    try:
        result = WorkersState.tasks[task['name']](*task['args'], **task['kwargs'])
        logging.info(
            'Task %s took %s seconds, and returned %s',
            formatted_task,
            (datetime.now() - task_start).total_seconds(),
            result,
        )
        return result
    except Exception:
        logging.info(
            'Task %s took %s seconds, and raised an exception:\n%s',
            formatted_task,
            (datetime.now() - task_start).total_seconds(),
            traceback.format_exc(),
        )
        remote_exc = RemoteException(traceback.format_exc())
        raise remote_exc


def scheduler_process(address, interval, new_schedule_queue):
    """Responsible for polling the tasks' schedules."""
    try:
        while True:
            start_time = datetime.now()
            while not new_schedule_queue.empty():
                new_schedule = new_schedule_queue.get(False)
                if new_schedule is None:
                    return
                fn_name, schedule = new_schedule
                WorkersState.schedules[fn_name].append(schedule)
            for fn_name, fn_schedules in WorkersState.schedules.copy().items():
                for loc, schedule in enumerate(fn_schedules[:]):
                    next_run, repeat_every, args, kwargs = schedule
                    if next_run < start_time:
                        if repeat_every:
                            fn_schedules[loc] = (next_run + repeat_every, repeat_every, args, kwargs)
                        else:
                            fn_schedules.pop(loc)
                        sendmsg(address, {'op': 'submit', 'task': {'name': fn_name, 'args': args, 'kwargs': kwargs}})
            sleep_duration = max(interval - (datetime.now() - start_time).total_seconds(), 0.1)
            time.sleep(sleep_duration)
    except KeyboardInterrupt:
        pass


def get_future_state(future):
    """Serialises the state of a concurrent.futures.Future"""
    if future is None:
        return {}
    with future._condition:
        return {
            'state': future._state,
            'result': future._result,
            'exception': future._exception,
        }


def check_for_chain(parent_id, executor, parent_future):
    """If there are tasks chained to the given parent, add them now."""
    with WorkersState.chains_lock:
        for task in WorkersState.chains[parent_id]:
            add_task(executor, task, parent_future)
        del WorkersState.chains[parent_id]


def future_values_fall_through(task, parent_future):
    """Pass the parent's state onto the child."""
    future = concurrent.futures.Future()
    future._state = parent_future._state
    future._result = parent_future._result
    future._exception = parent_future._exception
    future._task = task
    WorkersState.futures[task['id']] = future


def submit_new_task(executor, task):
    """Submit a new task to the given executor"""
    future = executor.submit(run_task, task)
    future.add_done_callback(functools.partial(check_for_chain, task['id'], executor))
    future._task = task
    WorkersState.futures[task['id']] = future


def add_task(executor, task, parent_future):
    """Add a new task to the executor, based on the result of the parent."""
    try:
        task['args'] = (parent_future.result(), ) + task['args']
    except BaseException as err:    # pylint: disable=W0703
        if task['type'] == 'catch':
            task['args'] = (err, ) + task['args']
            submit_new_task(executor, task)
        else:
            future_values_fall_through(task, parent_future)
    else:
        if task['type'] == 'then':
            submit_new_task(executor, task)
        else:
            future_values_fall_through(task, parent_future)


@contextlib.contextmanager
def create_workers(modules, num_workers):
    """Create the workers and yield the executor.

    .. doctest::

        >>> with create_workers([], 2) as workers:
        ...     pass

    :param modules: Strings representing the python paths of modules from which
        to load tasks.
    :type modules: list
    :param num_workers: The number of worker processes to spawn.
    :type num_workers: int
    :returns: The worker's executor.
    :rtype: concurrent.futures.Executor
    """

    logging.debug('Setting up process pool with %s workers', num_workers)
    executor = concurrent.futures.ProcessPoolExecutor(max_workers=num_workers)

    for mod in modules:
        logging.debug('Importing %s', mod)
        importlib.import_module(mod)

    try:
        yield executor
    except KeyboardInterrupt:
        print('KeyboardInterrupt')
    finally:
        logging.debug('Shutting down process pool and waiting for currently running tasks to finish')
        executor.shutdown(wait=True)


@contextlib.contextmanager
def setup_scheduler(address, interval):
    """Set up and start the scheduler, then yield.

    .. doctest::

        >>> with setup_scheduler(('localhost', 12378), 1.0):
        ...     pass

    :param address: The address on which to listen for connections.
    :type address: tuple
    :param interval: How often to poll the tasks' schedules, to see if a task
        should be run.
    :type interval: float
    """

    logging.debug('Setting up task scheduler thread')
    scheduler = threading.Thread(
        target=scheduler_process,
        args=(address, interval, WorkersState.schedule_queue),
        name='Scheduler',
    )
    scheduler.daemon = True

    logging.debug('Starting task scheduler')
    scheduler.start()
    try:
        yield
    except KeyboardInterrupt:
        print('KeyboardInterrupt')
    finally:
        logging.debug('Stopping task scheduler')
        WorkersState.schedule_queue.put(None)
        scheduler.join()


def get_future(task_id, client_address):
    """Get the future associated with the given task ID, and check that it was
    originally submitted by the same client.
    """
    future = WorkersState.futures[task_id]
    if future is not None:
        task = future._task
        if task['client'] != client_address:
            raise TaskWasNotSubmitted(task_id)
    return future


@contextlib.contextmanager
def setup_tcp_server(address, workers):
    """Set up and then yield the TCP server.

    :param address: The address on which to listen for connections.
    :type address: tuple
    :param wokrers: The workers with which to run the requested tasks.
    :type workers: concurrent.futures.Executor
    :returns: The TCP server.
    :rtype: socketserver.BaseRequestHandler
    """

    class TCPHandler(socketserver.BaseRequestHandler):
        """Handle requests for running tasks and getting results."""

        def _handle_task_submission(self, task):
            if task['name'] not in WorkersState.tasks:
                return TaskDoesNotExist(task['name'])
            task['client'] = self.client_address[0]
            task_id = str(uuid.uuid4()).encode()
            task['id'] = task_id
            logging.debug('Adding task %s to queue', format_task(task))
            submit_new_task(workers, task)
            return task

        def _handle_task_cancellation(self, task_id):
            future = get_future(task_id, self.client_address[0])
            if future is not None:
                response = future.cancel()
            else:
                response = False
                with WorkersState.chains_lock:
                    for children in WorkersState.chains.values():
                        for child in children[:]:
                            if child['id'] == task_id:
                                response = True
                                children.remove(child)
            if response:
                logging.debug('Cancelled task [%s]', task_id.decode())
            else:
                logging.debug('Failed to cancel task [%s]', task_id.decode())
            state = get_future_state(future)
            state['response'] = response
            return state

        def _handle_task_poll(self, task_id):
            future = get_future(task_id, self.client_address[0])
            state = get_future_state(future)
            logging.debug('Polled for task [%s]: %s', task_id.decode(), state)
            return state

        def _handle_task_chaining(self, task, parent_id):
            if task['name'] not in WorkersState.tasks:
                return TaskDoesNotExist(task['name'])
            task['client'] = self.client_address[0]
            task_id = str(uuid.uuid4()).encode()
            task['id'] = task_id
            future = get_future(parent_id, self.client_address[0])
            with WorkersState.chains_lock:
                if future is not None and future.done():
                    add_task(workers, task, future)
                else:
                    WorkersState.chains[parent_id].append(task)
                    WorkersState.futures[task_id] = None
            logging.debug('Chained task %s to [%s]', format_task(task, chain=True), parent_id.decode())
            return task

        def _handle_task_queue_inspection(self):
            task_list = {}
            for task_name, task_fn in WorkersState.tasks.items():
                try:
                    sig = signature(task_fn)
                except ValueError as err:
                    sig = err
                task_list[task_name] = str(sig)

            def _get_task_if_from_client(future):
                task = future._task
                addr = self.client_address[0]
                if task['client'] != addr and not addr.startswith('127.0.0'):
                    return None
                return task

            response = {
                'tasks': task_list,
                'schedules': WorkersState.schedules,
                'running': [_get_task_if_from_client(f) for f in WorkersState.futures.values() if f and f.running()],
            }
            return response

        def handle(self):
            self.data = self.request.recv(10240).strip()
            request = pickle.loads(self.data)
            logging.debug('Received request from %s: %s', self.client_address[0], request)
            # TODO: Check credentials
            # TODO: Encrypt credentials
            try:
                if request['op'] == 'submit':
                    task = request['task']
                    task = self._handle_task_submission(task)
                    self.request.sendall(pickle.dumps(task))
                elif request['op'] == 'cancel':
                    task_id = request['id']
                    state = self._handle_task_cancellation(task_id)
                    self.request.sendall(pickle.dumps(state))
                elif request['op'] == 'poll':
                    task_id = request['id']
                    state = self._handle_task_poll(task_id)
                    self.request.sendall(pickle.dumps(state))
                elif request['op'] == 'chain':
                    task = request['task']
                    parent_id = request['parent']
                    task = self._handle_task_chaining(task, parent_id)
                    self.request.sendall(pickle.dumps(task))
                elif request['op'] == 'inspect':
                    inspect_response = self._handle_task_queue_inspection()
                    self.request.sendall(pickle.dumps(inspect_response))
                else:
                    self.request.sendall(pickle.dumps(UnknownMessage(request)))
            except TaskWasNotSubmitted as error:
                self.request.sendall(pickle.dumps(error))
            except (KeyError, TypeError):
                self.request.sendall(pickle.dumps(UnknownMessage(request)))
            except Exception as error:
                logging.warning('Exception raised when processing request:\n%s', traceback.format_exc())
                self.request.sendall(pickle.dumps(error))

    logging.debug('Setting up TCP server')
    socketserver.ThreadingTCPServer.allow_reuse_address = True
    server = socketserver.ThreadingTCPServer(address, TCPHandler)

    try:
        yield server
    except KeyboardInterrupt:
        print('KeyboardInterrupt')
    finally:
        logging.debug('Shutting down TCP server')
        server.shutdown()
        server.server_close()



# -*- coding: utf-8 -*-

"""
.. module: inspect
    :synopsis: Inspect the state of the tasks and their schedules.
"""


__all__ = ('inspect_worker', 'print_inspected_worker')


from .client import sendmsg
from .utils import format_fn


def inspect_worker(address):
    """Return a dict containing the state of the tasks and their schedules."""
    return sendmsg(address, {'op': 'inspect'})


def _print_singly_scheduled_tasks(tasks_and_schedules):
    print('Scheduled tasks:')
    for task_name, schedules in tasks_and_schedules.items():
        for schedule in schedules:
            if not schedule[1]:
                print('  * {}:\tScheduled for {}'.format(
                    format_fn(task_name, schedule[2], schedule[3]),
                    schedule[0].isoformat(' '),
                ))
    print('')


def _print_repeatedly_scheduled_tasks(tasks_and_schedules):
    print('Recurring tasks:')
    for task_name, schedules in tasks_and_schedules.items():
        for schedule in schedules:
            if schedule[1]:
                print('  * {}:\tScheduled for {}; repeating every {}'.format(
                    format_fn(task_name, schedule[2], schedule[3]),
                    schedule[0].isoformat(' '),
                    str(schedule[1]),
                ))
    print('')


def print_inspected_worker(details):
    """Print the state of the tasks and their schedules to the console."""
    if details['tasks']:
        print('Registered tasks:')
        for task_name, task_sig in details['tasks'].items():
            print('  * {}{}'.format(task_name, task_sig))
        print('')

    if details['running']:
        print('Running tasks:')
        for task in details['running']:
            if task is not None:
                task_id = task['id'].decode()
                print('  * {}:   {}'.format(task_id, format_fn(task['name'], task['args'], task['kwargs'])))
        print('')

    if details['schedules']:
        single_sched = {t: s for t, s in details['schedules'].items() if [sch for sch in s if not sch[1]]}
        repeat_sched = {t: s for t, s in details['schedules'].items() if [sch for sch in s if sch[1]]}

        if single_sched:
            _print_singly_scheduled_tasks(single_sched)

        if repeat_sched:
            _print_repeatedly_scheduled_tasks(repeat_sched)



import functools
import pytest
import tempfile
import time

from datetime import datetime, timedelta

import apium

from apium.inspect import inspect_worker, print_inspected_worker


def send_msg_from(address, port_num, bind_addr, msg):
    import socket
    import pickle
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind((bind_addr, 0))
    try:
        sock.connect((address, port_num))
        sock.sendall(pickle.dumps(msg))
        received = sock.recv(10240)
        result = pickle.loads(received)
        if isinstance(result, Exception):
            raise result
        return result
    finally:
        sock.close()


def test_basic_task_run___state_is_consistent(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('add', 1, 2, 3)
        time.sleep(0.1)
        assert task.done() is False
        assert task.running() is True
        assert task.result() is 6
        assert task.done() is True
        assert task.running() is False


def test_mapping_tasks___values_are_returned_as_tasks_complete(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        results = list(executor.map('add', (2, 4, 6), (3, 6, 9)))
        assert 5 in results
        assert 10 in results
        assert 15 in results
        assert len(results) is 3


def test_cancelling_task___functions_like_a_standard_future(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        values = list(range(8))
        tasks = [executor.submit('add', value) for value in values]
        assert tasks[-1].cancel() is True
        assert tasks[-1].done() is True
        assert tasks[0].result() in values
        assert tasks[0].cancel() is False
        with pytest.raises(apium.CancelledError):
            tasks[-1].result()


def test_chaining_futures___results_of_previous_tasks_are_passed_forward(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('add', 1, 1).then('add', 2).then('add', 4).then('add', 8)
        assert task.done() is False
        assert task.running() is False
        assert task.result() is 16
        assert task.done() is True
        assert task.running() is False


def test_chaining_multiple_times_from_one_future___functions_like_a_single_chain(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        initial_task = executor.submit('add', 1, 1).then('add', 2)
        chain1 = initial_task.then('add', 5)
        chain2 = initial_task.then('add', 7)
        assert chain1.done() is chain2.done() is False
        assert chain1.running() is chain2.running() is False
        assert chain1.result() is 9
        assert chain2.result() is 11
        assert chain1.done() is chain2.done() is True
        assert chain1.running() is chain2.running() is False


def test_cancelling_chained_task___functions_as_a_standard_future(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('add', 1, 1).then('add', 2).then('add', 4).then('add', 8)
        assert task.done() is False
        assert task.running() is False
        assert task.cancel() is True
        assert task.done() is True
        with pytest.raises(apium.CancelledError):
            task.result()


def test_tasks_raising_exception___exception_raised_locally_with_stacktrace(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('raiser')
        with pytest.raises(apium.RemoteException):
            try:
                task.result()
            except Exception as err:
                assert 'ValueError' in str(err)
                raise


def test_tasks_raising_exception_in_chain___exception_propagated(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('add', 1, 1).then('add', 2).then('raiser')
        with pytest.raises(apium.RemoteException):
            task.result()


def test_tasks_catching_exception_in_chain___exception_not_propagated(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('raiser').catch('format_exc')
        assert 'ValueError' in task.result()


def test_tasks_exceptions_falling_through_tasks___exception_propagated(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('raiser').then('add', 2)
        with pytest.raises(apium.RemoteException):
            task.result()


def test_tasks_results_falling_through_catch___result_propagated(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('add', 2, 3).catch('format_exc')
        assert task.result() is 5


def test_tasks_chaining_on_a_finished_task___chaining_as_per_normal(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('add', 2, 3)
        task.result()
        chain = task.then('add', 5)
        assert chain.result() is 10


def test_importing_tasks_from_module___tasks_can_be_run(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('chain', [1, 2, 3], [4, 5, 6])
        assert list(task.result()) == [1, 2, 3, 4, 5, 6]


def test_calling_non_existant_task___exception_raised(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        with pytest.raises(apium.TaskDoesNotExist):
            executor.submit('non-existant')


def test_chaining_non_existant_task___exception_raised(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        with pytest.raises(apium.TaskDoesNotExist):
            executor.submit('add', 2, 3).then('non-existant')


def test_chaining_task_that_is_yet_to_be_queued___chaining_as_per_normal(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        values = list(range(6))
        tasks = [executor.submit('add', value) for value in values]
        chained_task = tasks[-1].then('add', 5)
        assert chained_task.result() is 10


def test_task_timing_out___exception_raised(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('add')
        with pytest.raises(apium.TaskTimeoutError):
            task.result(timeout=0.01)


def test_scheduling_tasks___tasks_called_on_schedule(port_num, running_worker):
    with tempfile.NamedTemporaryFile() as tmpfile:
        filename = tmpfile.name

        wait = 0.1
        import task_import
        apium.schedule_task(
            datetime.now() + timedelta(seconds=wait),
            args=(filename, )
        )(task_import.scheduled_fn)
        time.sleep(3 * wait)
        assert len(open(filename).read()) is 1


def test_scheduling_repeating_tasks___tasks_called_on_repeating_schedule(port_num, running_worker):
    with tempfile.NamedTemporaryFile() as tmpfile:
        filename = tmpfile.name

        wait = 0.1
        import task_import
        apium.schedule_task(
            datetime.now(),
            timedelta(seconds=wait),
            args=(filename, )
        )(task_import.scheduled_fn)
        time.sleep(5 * wait)
        assert len(open(filename).read()) > 3


def test_sending_a_garbage_message___exception_raised1(port_num, running_worker):
    with pytest.raises(apium.UnknownMessage):
        apium.client.sendmsg(('localhost', port_num), {'op': 'bob'})


def test_sending_a_garbage_message___exception_raised2(port_num, running_worker):
    with pytest.raises(apium.UnknownMessage):
        apium.client.sendmsg(('localhost', port_num), 'bob')


def test_cancelling_task_after_executor_dies___exception_raised(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('add', 1, 2, 3)
    with pytest.raises(apium.DeadExecutor):
        task.cancel()


def test_chaininging_task_after_executor_dies___exception_raised(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        task = executor.submit('add', 1, 2, 3)
    with pytest.raises(apium.DeadExecutor):
        task.then('add', 4)


def test_submitting_task_after_executor_shutdown___exception_raised(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        executor._shutting_down = True
        with pytest.raises(RuntimeError):
            executor.submit('add', 1, 2, 3)


def test_cancelling_task___executor_doesnt_wait_for_cancelled_task_on_shutdown(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        values = list(range(8))
        tasks = [executor.submit('add', value) for value in values]
        assert tasks[-1].cancel() is True


def test_shutting_down_executor_without_waiting___executor_shuts_down_cleanly(port_num, running_worker):
    executor = apium.TaskExecutor(port=port_num, polling_interval=0.1)
    values = list(range(8))
    for value in values:
        executor.submit('add', value)
    executor.shutdown(wait=False)


def test_wait_all_completed___all_futures_done(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        values = list(range(6))
        tasks = [executor.submit('add', value) for value in values]
        results = apium.wait(tasks)
        assert len(results.done) == len(values)
        assert len(results.not_done) == 0


def test_wait_first_completed___one_future_done(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        values = list(range(6))
        tasks = [executor.submit('add', value) for value in values]
        results = apium.wait(tasks, return_when=apium.FIRST_COMPLETED)
        assert len(results.done) >= 1
        assert len(results.not_done) <= len(values) - 1


def test_wait_first_exception___all_futures_done(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.02) as executor:
        tasks = [executor.submit('add', 1)]
        tasks.append(executor.submit('raiser'))
        tasks += [executor.submit('add', 2)]
        results = apium.wait(tasks, return_when=apium.FIRST_EXCEPTION)
        assert len(results.done) == 1
        assert len(results.not_done) == 2


def test_as_completed___futures_returned_as_sompleted(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.1) as executor:
        values = list(range(6))
        tasks = list(apium.as_completed([executor.submit('add', value) for value in values]))
        assert len(tasks) == len(values)
        assert functools.reduce(lambda d1, d2: d1 and d2, map(lambda f: f.done(), tasks))
        assert sum(map(lambda f: f.result(), tasks)) == sum(values)


def test_inspect___registered_task_names_returned(port_num, running_worker):
    details = inspect_worker(('localhost', port_num))
    tasks = details['tasks']
    assert len(tasks) == 5
    task_names = tasks.keys()
    assert 'chain' in task_names
    assert 'add' in task_names
    assert 'raiser' in task_names
    assert 'scheduled_fn' in task_names
    assert 'format_exc' in task_names


def test_inspect___registered_signature_returned(port_num, running_worker):
    details = inspect_worker(('localhost', port_num))
    tasks = details['tasks']
    assert len(tasks) == 5
    assert len([t for t in tasks.values() if isinstance(t, str)]) == 5


def test_inspect___schedules_of_tasks_returned_with_correct_repeat(port_num, running_worker):
    details = inspect_worker(('localhost', port_num))
    schedules = details['schedules']
    assert len(schedules) == 1
    assert len(schedules['scheduled_fn']) == 1
    assert schedules['scheduled_fn'][0][1] == timedelta(seconds=0.1)


def test_inspect___schedules_of_tasks_returned_with_correct_initial(port_num, running_worker):
    future_initial = datetime.now() + timedelta(seconds=3600)
    import task_import
    apium.schedule_task(
        future_initial,
        args=('/tmp/a/b/c/d/e/f', ),
        kwargs={'fake': 'arg'},
    )(task_import.scheduled_fn)
    time.sleep(0.2)

    details = inspect_worker(('localhost', port_num))
    schedules = details['schedules']
    assert len(schedules) == 1
    assert len(schedules['scheduled_fn']) == 2
    assert schedules['scheduled_fn'][1][0] == future_initial
    assert schedules['scheduled_fn'][1][2] == ('/tmp/a/b/c/d/e/f', )
    assert schedules['scheduled_fn'][1][3] == {'fake': 'arg'}


def test_print_inspect___runs_without_exception(port_num, running_worker):
    with apium.TaskExecutor(port=port_num, polling_interval=0.02) as executor:
        executor.submit('add', 1)
        print_inspected_worker(inspect_worker(('localhost', port_num)))


def test_framework_setup_without_extra_frameworks___happily_continues(port_num, running_worker):
    import apium.frameworks
    apium.frameworks.setup()


def test_tasks_are_segregated_by_address___cant_check_task_from_different_ip(port_num, running_worker):
    with apium.TaskExecutor(server='127.0.0.1', port=port_num, polling_interval=0.1) as executor:
        future1 = executor.submit('add', 2, 3)
        send_msg_from('127.0.0.1', port_num, '127.0.0.1', {'op': 'poll', 'id': future1._id})
        future2 = executor.submit('add', 2, 3)
        with pytest.raises(apium.TaskWasNotSubmitted):
            send_msg_from('127.0.0.1', port_num, '127.0.0.3', {'op': 'poll', 'id': future2._id})


def test_inspect_from_other_client___cant_see_running_tasks(port_num, running_worker):
    with apium.TaskExecutor(server='127.0.0.1', port=port_num, polling_interval=0.1) as executor:
        while len(inspect_worker(('localhost', port_num))['running']) > 0:
            time.sleep(0.1)
        executor.submit('add', 2, 3)
        executor.submit('add', 2, 3)
        details = send_msg_from('127.0.0.1', port_num, '127.0.0.1', {'op': 'inspect'})
        assert len(details['running']) == 2
        assert details['running'][0] is not None
        assert details['running'][1] is not None
        details = send_msg_from('127.0.0.1', port_num, '127.1.1.3', {'op': 'inspect'})
        assert len(details['running']) == 2
        assert details['running'][0] is None
        assert details['running'][1] is None



from apium.crypt.dh import DiffieHellman
DiffieHellman.p = 0x7fffffff
DiffieHellman.g = 2

from apium.crypt import Encryptor


def test_diffie_hellman___generated_keys_are_the_same(port_num, running_worker):
    alice = DiffieHellman()
    bob = DiffieHellman()

    alice_key = alice.generate_key(bob.public_key)
    bob_key = bob.generate_key(alice.public_key)

    assert alice_key == bob_key


def test_message_encryption___message_could_be_decrypted(port_num, running_worker):
    server_enc = Encryptor()
    client_enc = Encryptor(server_enc._pk)

    msg = {'a': 'Allo', 'b': 2}
    encrypted_msg = client_enc.encrypt(msg)
    assert msg != encrypted_msg['data']

    decrypted_msg = server_enc.decrypt(encrypted_msg)
    assert msg == decrypted_msg


def test_message_encryption_when_libraries_not_installed___messages_assed_through(port_num, running_worker):
    server_enc = Encryptor()
    client_enc = Encryptor(server_enc._pk)
    client_enc._fern = 0
    client_enc._pk = None

    msg = {'a': 'Allo', 'b': 2}
    encrypted_msg = client_enc.encrypt(msg)
    assert msg == encrypted_msg['data']

    decrypted_msg = server_enc.decrypt(encrypted_msg)
    assert msg == decrypted_msg



# -*- coding: utf-8 -*-

"""
.. module: dh
    :synopsis: Simple implementation of Diffie-Hellman.
"""


__all__ = ('DiffieHellman', )


import hashlib
from random import randint
from binascii import hexlify


class DiffieHellman(object):
    """A simple implementation of Diffie-Hellman for key-sharing."""

    # This prime comes from: http://tools.ietf.org/html/rfc3526#section-5
    modulus_p = 0xFFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD129024E088A67CC74020BBEA63B139B22514A08798E3404DDEF9519B3CD3A431B302B0A6DF25F14374FE1356D6D51C245E485B576625E7EC6F44C42E9A637ED6B0BFF5CB6F406B7EDEE386BFB5A899FA5AE9F24117C4B1FE649286651ECE45B3DC2007CB8A163BF0598DA48361C55D39A69163FA8FD24CF5F83655D23DCA3AD961C62F356208552BB9ED529077096966D670C354E4ABC9804F1746C08CA18217C32905E462E36CE3BE39E772C180E86039B2783A2EC07A28FB5C55DF06F4C52C9DE2BCBF6955817183995497CEA956AE515D2261898FA051015728E5A8AAAC42DAD33170D04507A33A85521ABDF1CBA64ECFB850458DBEF0A8AEA71575D060C7DB3970F85A6E1E4C7ABF5AE8CDB0933D71E8C94E04A25619DCEE3D2261AD2EE6BF12FFA06D98A0864D87602733EC86A64521F2B18177B200CBBE117577A615D6C770988C0BAD946E208E24FA074E5AB3143DB5BFCE0FD108E4B82D120A92108011A723C12A787E6D788719A10BDBA5B2699C327186AF4E23C1A946834B6150BDA2583E9CA2AD44CE8DBBBC2DB04DE8EF92E8EFC141FBECAA6287C59474E6BC05D99B2964FA090C3A2233BA186515BE7ED1F612970CEE2D7AFB81BDD762170481CD0069127D5B05AA993B4EA988D8FDDC186FFB7DC90A6C08F4DF435C934063199FFFFFFFFFFFFFFFF  # pylint: disable=C0301
    base_g = 2

    def __init__(self):
        self.private_key = self._generate_primary_key()
        self.public_key = self._generate_public_key()

    def _generate_primary_key(self):
        return randint(2, self.modulus_p - 1)

    def _generate_public_key(self):
        return pow(self.base_g, self.private_key, self.modulus_p)

    def generate_key(self, another_key):
        """Generate a secret key based on the public key of another instance.

        :param another_key: The public key of another DiffieHellman instance.
        :type another_key: int
        :returns: A key only decipherable by the originator of 'another_key'.
        :rtype: str
        """
        shared_secret = pow(another_key, self.private_key, self.modulus_p)
        hashed_secret = hashlib.sha256()
        hashed_secret.update(str(shared_secret).encode())
        key = hashed_secret.digest()
        return hexlify(key)



# -*- coding: utf-8 -*-

"""
.. module: client
    :synopsis: Basic functions for interacting with the server.
"""


__all__ = ('sendmsg', )


import pickle
import socket


DEFAULT_PORT = 9737


def sendmsg(address, data):
    """Send the data to the given address.

    The data will be pickled before sending it, so the given data needs
    to be pickleable.

    :param address: The location and port of the server.
    :type address: tuple
    :returns: The server's response
    :rtype: object
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect(address)
        sock.sendall(pickle.dumps(data))
        received = sock.recv(10240)
        result = pickle.loads(received)
        if isinstance(result, Exception):
            raise result
        return result
    finally:
        sock.close()



# -*- coding: utf-8 -*-

"""
.. module: crypt
    :synopsis: Message encryption and decryption.
"""


__all__ = ('Encryptor', )


import base64
import pickle

from .dh import DiffieHellman


class NoEncryptionLibrariesInstalled(Exception):
    """Raised if an encrypted message is received, but there are no libraries
    available to decrypt the message.
    """
    pass


class Encryptor:
    """Handles the encryption and decryption of messages.

    Note that this requires the 'cryptography' package to be installed,
    otherwise this class will default back to unencrypted messages.

    Here's a brief example of how to use the Encryptors:

    .. doctest::

        >>> from apium.crypt import Encryptor
        >>> server_enc = Encryptor()
        >>> client_enc = Encryptor(server_enc._pk)
        >>> msg = {'a': 'Allo', 'b': 2}
        >>> encrypted_msg = client_enc.encrypt(msg)
        >>> msg != encrypted_msg['data']
        True
        >>> decrypted_msg = server_enc.decrypt(encrypted_msg)
        >>> msg == decrypted_msg
        True

    :param remote_pk: The public key from another Encryptor instance.
    :type remote_pk: int
    """

    def __init__(self, remote_pk=None):
        self._fern = None
        self._dh = DiffieHellman()
        self._pk = self._dh.public_key
        self._remote_pk = remote_pk

    def _fernet(self):
        if self._fern is None:
            try:
                from cryptography.fernet import Fernet
                key = self._dh.generate_key(self._remote_pk)
                self._fern = Fernet(base64.urlsafe_b64encode(key[:32]))
            except ImportError:
                self._pk = None
                self._fern = 0
        return self._fern

    def encrypt(self, msg):
        """Return a dict containing the (pickled then) encrypted python object.

        Note that this method requires access to the public key of the Encryptor
        that will be decrypting this message, either by having received a
        message from it previously, or by explicitly passing it in during
        construction.

        :param msg: A pickleable python object.
        :type msg: object
        :returns: The encrypted message format to pass to another Encryptor.
        :rtype: dict
        """
        return {
            'data': self._fernet().encrypt(pickle.dumps(msg)) if self._fernet() else msg,
            'pk': self._pk,
        }

    def decrypt(self, msg):
        """Return the decrypted message.

        :param msg: A dict returned by encrypt()
        :type msg: dict
        :returns: The original unencrypted message from the originating
            Encryptor.
        :rtype: object
        """
        remote_pk = msg['pk']
        if remote_pk:
            self._remote_pk = remote_pk
            if not self._fernet():
                raise NoEncryptionLibrariesInstalled('"cryptography" not installed - cannot decrypt message')
            return pickle.loads(self._fernet().decrypt(msg['data']))
        else:
            return msg['data']



# -*- coding: utf-8 -*-

"""
.. module: utils
    :synopsis: A collection of useful, cross-module functions.
"""


__all__ = ('format_fn', )


def format_fn(name, args, kwargs):
    """Return a string formatting the given function name, args and kwargs.

    Outputted string resembles a function call in python:

    .. doctest::

        >>> format_fn('fn_name', (1, 2, 'c'), {'some': 24, 'another': 'blah'})
        "fn_name(1, 2, 'c', some=24, another='blah')"

    :param name: The name of the function.
    :type name: str
    :param args: The arguments to pass to the function.
    :type args: tuple
    :param kwargs: The keyword arguments to pass to the function.
    :type kwargs: dict
    :returns: A string representation of a function call.
    :rtype: str
    """

    def _format_args():
        return ', '.join([repr(a) for a in args or ()])

    def _format_kwargs():
        return ', '.join([str(k) + '=' + repr(v) for k, v in (kwargs or {}).items()])

    return '{}({}{}{})'.format(
        name,
        _format_args(),
        ', ' if args and kwargs else '',
        _format_kwargs(),
    )



# -*- coding: utf-8 -*-

"""
.. module: frameworks
    :synopsis: Functions to ease integration with various frameworks.
"""


__all__ = ('setup', )


import logging
import os


DJANGO_SETTINGS_ENV_VAR = 'DJANGO_SETTINGS_MODULE'


def _setup_django():
    try:
        import django
        django.setup()
        from django.core.checks import run_checks
        run_checks()
    except ImportError as error:
        logging.error('Found %s in environment, but couldn\'t import django: %s', DJANGO_SETTINGS_ENV_VAR, error)
        raise


def setup():
    """Check if any frameworks appear to be in use, and set them up."""
    if DJANGO_SETTINGS_ENV_VAR in os.environ:
        _setup_django()
